--------------
-- Union bench
--------------

-- BTree
data BTree a = BLeaf a | BBranch (BTree a) (BTree a) deriving (Show)
instance NFData (BTree a) where
  rnf (BLeaf a) = ()
  rnf (BBranch left right) = case rnf left of () -> rnf right

balBTree 0 (e:es) = (BLeaf e, es)
balBTree levels es = levels `seq`
  case balBTree (levels-1) es of
    (bLeft, esLeft) -> case balBTree (levels-1) esLeft of
      (bRight, esRight) -> (BBranch bLeft bRight, esRight)

unbalBTree levels es = unbal' (2^levels-1) es
  where unbal' 1 (e:es) = (BLeaf e, es)
        unbal' n es = let nLeft = (n - 1) `div` 6 + 1
                      in case unbal' nLeft es of
                        (bLeft, esLeft) -> case unbal' (n-nLeft) esLeft of
                          (bRight, esRight) -> (BBranch bLeft bRight, esRight)

mirror (BLeaf a) = BLeaf a
mirror (BBranch left right) = BBranch (mirror left) (mirror right)

pathBTree levels es = path' (2^levels-1) es
  where path' 1 (e:es) = (BLeaf e, es)
        path' n (e:es) = case path' (n-1) es of
                           (bRight, esRight) -> (BBranch (BLeaf e) bRight, esRight)

bAsc level fun = fst $ fun level [(1::Int)..]
bDsc level fun = fst $ fun level [(2^31-1::Int),2^31-2..]
-- bRnd level fun = fst $ fun level (randomRs (0,100000000) $ mkStdGen 1234)
bRnd level fun = fst $ fun level list
  where list' = take (2^level) $ randomRs (0,100000000) $ mkStdGen 1234
        list = seq (rnf list') list'

-- Union using BTree
bTreeUnion (BLeaf a) = sSingleton a
bTreeUnion (BBranch bLeft bRight) =
  case bTreeUnion bLeft of
    fmLeft -> case bTreeUnion bRight of
      fmRight -> fmLeft `sUnion` fmRight

-- The benchmark
benchUnion k = case [fun k balBTree | fun <- [bAsc, bRnd]] of
  [asc_b, rnd_b]-> case [fun k unbalBTree | fun <- [bAsc, bRnd]] of
   [asc_r, rnd_r] -> case map mirror [asc_r, rnd_r] of
    [asc_l, rnd_l] -> case [fun k pathBTree | fun <- [bAsc, bRnd]] of
     [asc_pr, rnd_pr] -> case map mirror [asc_pr, rnd_pr] of
      [asc_pl, rnd_pl] -> case map rnf [asc_b,rnd_b,asc_l,rnd_l,asc_r,rnd_r,asc_pr,rnd_pr,asc_pl,rnd_pl] of
       [(), (), (), (), (), (), (), (), (), ()] ->
        Criterion.bgroup (num02 k) [
          Criterion.bench "asc_b" $ Criterion.whnf (sForce . bTreeUnion) asc_b,
          Criterion.bench "asc_u" $ Criterion.whnf (sForce . bTreeUnion) asc_l,
--          Criterion.bench "asc_ur" $ Criterion.whnf (sForce . bTreeUnion) asc_r,
--          Criterion.bench "asc_pr" $ Criterion.whnf (sForce . bTreeUnion) asc_pr,
          Criterion.bench "asc_p" $ Criterion.whnf (sForce . bTreeUnion) asc_pl,
          Criterion.bench "rnd_b" $ Criterion.whnf (sForce . bTreeUnion) rnd_b,
          Criterion.bench "rnd_u" $ Criterion.whnf (sForce . bTreeUnion) rnd_l,
--          Criterion.bench "rnd_ur" $ Criterion.whnf (sForce . bTreeUnion) rnd_r,
          Criterion.bench "rnd_p" $ Criterion.whnf (sForce . bTreeUnion) rnd_pr
--          Criterion.bench "rnd_pl" $ Criterion.whnf (sForce . bTreeUnion) rnd_pl
        ]

benchUnions lens = Criterion.bgroup "treeunion" [benchUnion l | l <- lens]

---------------------
-- Big union bench --
---------------------

benchBigUnion (s,b) = case map bInserts [asc_ls, asc_lb, e_o_ls, e_o_lb, mix_ls, mix_lb, rnd_ls, rnd_lb] of
  all@[asc_s, asc_b, e_o_s, e_o_b, mix_s, mix_b, rnd_s, rnd_b] ->
    Criterion.bgroup (num02 s ++ "/" ++ num02 b) $ [
      Criterion.bench "asc" $ Criterion.whnf (sForce . (uncurry sUnion)) (asc_s, asc_b),
      Criterion.bench "e_o" $ Criterion.whnf (sForce . (uncurry sUnion)) (e_o_s, e_o_b),
      Criterion.bench "mix" $ Criterion.whnf (sForce . (uncurry sUnion)) (mix_s, mix_b)
    ]
  where sl = 2^s - 1
        bl = 2^b - 1
        asc_ls = [1..sl]
        asc_lb = [sl+1..sl+bl]

        e_o_ls = [1,3..2*sl]
        e_o_lb = [2,4..2*bl]

        rnd_ls = take sl (randomRs (0,100 * sl) $ mkStdGen 1234)
        rnd_lb = take bl (randomRs (0,100 * bl) $ mkStdGen 1234)

        (mix_ls, mix_lb) = split [1..] sl (2^(s`div`2)) bl (2^(b-s`div`2))
        split xs aleft alen bleft blen | aleft <= alen || bleft <= blen =
          let (a, xs') = splitAt alen xs
          in (a, take blen xs')
                                       | otherwise =
          let (a, xs') = splitAt alen xs
              (b, xs'') = splitAt blen xs'
              (a', b') = split xs'' (aleft-alen) alen (bleft-blen) blen
          in (a++a', b++b')


benchBigUnions lens = Criterion.bgroup "union" [benchBigUnion l | l <- lens]

------------------
-- Insert bench --
------------------
bInserts (x:xs) = foldl' sInsert (sSingleton x) xs

lAsc, lDsc, lRnd :: Int -> [Int]
lAsc len = [1..len]
lDsc len = [len,len-1..1]
lRnd len = take len (randomRs (0,100 * len) $ mkStdGen 1234)

benchInsert k = case [lAsc (2^k), lRnd (2^k)] of
  [asc_l, rnd_l] -> case map rnf [asc_l, rnd_l] of
    [(), ()] ->
      Criterion.bgroup (num02 k) [
        Criterion.bench "asc" $ Criterion.whnf (sForce . bInserts) asc_l,
        Criterion.bench "rnd" $ Criterion.whnf (sForce . bInserts) rnd_l
      ]

benchInserts lens = Criterion.bgroup "insert" [benchInsert l | l <- lens]

------------------
-- Delete bench --
------------------
bDeletes s xs = foldl' sDelete s xs

benchDelete k = case [lAsc (2^k), lRnd (2^k)] of
  [asc_l, rnd_l] -> case map bInserts [asc_l, rnd_l] of
    [asc_fm, rnd_fm] ->
      Criterion.bgroup (num02 k) [
        Criterion.bench "asc" $ Criterion.whnf (sForce . bDeletes asc_fm) asc_l,
        Criterion.bench "rnd" $ Criterion.whnf (sForce . bDeletes rnd_fm) rnd_l
      ]

benchDeletes lens = Criterion.bgroup "delete" [benchDelete l | l <- lens]

------------------
-- Lookup bench --
------------------
bLookups list map = add list map 0
  where add _ _ acc | acc `seq` False = undefined
        add [] map acc = acc
        add (e:es) map acc | map `sLookup` e = add es map (acc + 1)
                           | otherwise       = add es map acc

benchLookup k = case [lAsc (2^k), lRnd (2^k)] of
  [asc_l, rnd_l] -> case map rnf [asc_l, rnd_l] of
   [(), ()] -> case map bInserts [asc_l, rnd_l] of
    [asc_fm, rnd_fm] ->
      Criterion.bgroup (num02 k) [
        Criterion.bench "asc" $ Criterion.whnf (bLookups asc_l) asc_fm,
        Criterion.bench "rnd" $ Criterion.whnf (bLookups rnd_l) rnd_fm
      ]

benchLookups lens = Criterion.bgroup "lookup" [benchLookup l | l <- lens]

-------------
-- All benchs
-------------
num02 n | n < 10 = '0' : (show n)
        | otherwise = (show n)

main = do
  args <- getArgs
  let benchname = args !! (fromJust (elemIndex "-p" args) + 1)
  Progression.defaultMain $ Criterion.bgroup "" $ case benchname of
    "treeunion" -> [ benchUnions [8, 14, 20] ]
    "union" -> [ benchBigUnions [(5, 10), (10, 10), (10, 20), (20, 20)] ]
    "insert" -> [ benchInserts [4, 8, 12, 16, 20] ]
    "delete" -> [ benchDeletes [4, 8, 12, 16, 20] ]
    "lookup" -> [ benchLookups [4, 8, 12, 16, 20] ]
